package com.tzld.piaoquan.recommend.server.service.rank.strategy;

import com.ctrip.framework.apollo.spring.annotation.ApolloJsonValue;
import com.google.common.reflect.TypeToken;
import com.tzld.piaoquan.recommend.server.common.base.RankItem;
import com.tzld.piaoquan.recommend.server.model.Video;
import com.tzld.piaoquan.recommend.server.service.rank.RankParam;
import com.tzld.piaoquan.recommend.server.service.rank.RankResult;
import com.tzld.piaoquan.recommend.server.service.rank.RankService;
import com.tzld.piaoquan.recommend.server.service.rank.extractor.ExtractorUtils;
import com.tzld.piaoquan.recommend.server.service.rank.extractor.RankExtractorItemTags;
import com.tzld.piaoquan.recommend.server.service.rank.processor.RankProcessorBoost;
import com.tzld.piaoquan.recommend.server.service.rank.processor.RankProcessorDensity;
import com.tzld.piaoquan.recommend.server.service.rank.processor.RankProcessorInsert;
import com.tzld.piaoquan.recommend.server.service.rank.processor.RankProcessorTagFilter;
import com.tzld.piaoquan.recommend.server.util.CommonCollectionUtils;
import com.tzld.piaoquan.recommend.server.util.JSONUtils;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.lang3.RandomUtils;
import org.springframework.stereotype.Service;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.*;
import java.util.stream.Collectors;

/**
 * @author zhangbo
 * @desc 地域召回融合 流量池汤姆森
 */
@Service
@Slf4j
public abstract class RankStrategy4RegionMergeModelBasic extends RankService {
    @ApolloJsonValue("${RankStrategy4DensityFilterV2:}")
    private Map<String, Map<String, Map<String, String>>> filterRules = new HashMap<>();
    String CLASS_NAME = this.getClass().getSimpleName();

    public void duplicate(Set<Long> setVideo, List<Video> videos) {
        Iterator<Video> iterator = videos.iterator();
        while (iterator.hasNext()) {
            Video v = iterator.next();
            if (setVideo.contains(v.getVideoId())) {
                iterator.remove();
            } else {
                setVideo.add(v.getVideoId());
            }
        }
    }

    @Override
    public RankResult mergeAndSort(RankParam param, List<Video> rovVideos, List<Video> flowVideos) {

        //1 兜底策略，rov池子不足时，用冷启池填补。直接返回。
        if (CollectionUtils.isEmpty(rovVideos)) {
            if (param.getSize() < flowVideos.size()) {
                return new RankResult(flowVideos.subList(0, param.getSize()));
            } else {
                return new RankResult(flowVideos);
            }
        }

        //2 根据实验号解析阿波罗参数。
        Set<String> abExpCodes = param.getAbExpCodes();
        Map<String, Map<String, String>> rulesMap = Collections.emptyMap();
        if (CollectionUtils.isNotEmpty(abExpCodes)) {
            for (Map.Entry<String, Map<String, Map<String, String>>> entry : this.filterRules.entrySet()) {
                if (abExpCodes.contains(entry.getKey())) {
                    rulesMap = entry.getValue();
                    break;
                }
            }
        }


        //3 标签读取
        if (rulesMap != null && !rulesMap.isEmpty()) {
            RankExtractorItemTags extractorItemTags = new RankExtractorItemTags(this.redisTemplate);
            extractorItemTags.processor(rovVideos, flowVideos);
        }
        //6 合并结果时间卡控
        if (rulesMap != null && !rulesMap.isEmpty()) {
            RankProcessorTagFilter.processor(rovVideos, flowVideos, rulesMap);
        }

        //4 rov池提权功能
        RankProcessorBoost.boostByTag(rovVideos, rulesMap);

        //5 rov池强插功能
        RankProcessorInsert.insertByTag(param, rovVideos, rulesMap);

        //7 流量池按比例强插
        List<Video> result = new ArrayList<>();
        for (int i = 0; i < param.getTopK() && i < rovVideos.size(); i++) {
            result.add(rovVideos.get(i));
        }
        double flowPoolP = getFlowPoolP(param);
        int flowPoolIndex = 0;
        int rovPoolIndex = param.getTopK();
        for (int i = 0; i < param.getSize() - param.getTopK(); i++) {
            double rand = RandomUtils.nextDouble(0, 1);
            if (rand < flowPoolP) {
                if (flowPoolIndex < flowVideos.size()) {
                    result.add(flowVideos.get(flowPoolIndex++));
                } else {
                    break;
                }
            } else {
                if (rovPoolIndex < rovVideos.size()) {
                    result.add(rovVideos.get(rovPoolIndex++));
                } else {
                    break;
                }
            }
        }
        if (rovPoolIndex >= rovVideos.size()) {
            for (int i = flowPoolIndex; i < flowVideos.size() && result.size() < param.getSize(); i++) {
                result.add(flowVideos.get(i));
            }
        }
        if (flowPoolIndex >= flowVideos.size()) {
            for (int i = rovPoolIndex; i < rovVideos.size() && result.size() < param.getSize(); i++) {
                result.add(rovVideos.get(i));
            }
        }

        //8 合并结果密度控制
        Map<String, Integer> densityRules = new HashMap<>();
        if (rulesMap != null && !rulesMap.isEmpty()) {
            for (Map.Entry<String, Map<String, String>> entry : rulesMap.entrySet()) {
                String key = entry.getKey();
                Map<String, String> value = entry.getValue();
                if (value.containsKey("density")) {
                    densityRules.put(key, Integer.valueOf(value.get("density")));
                }
            }
        }
        Set<Long> videosSet = result.stream().map(Video::getVideoId).collect(Collectors.toSet());
        List<Video> rovRecallRankNew = rovVideos.stream().filter(r -> !videosSet.contains(r.getVideoId())).collect(Collectors.toList());
        List<Video> flowPoolRankNew = flowVideos.stream().filter(r -> !videosSet.contains(r.getVideoId())).collect(Collectors.toList());
        List<Video> resultWithDensity = RankProcessorDensity.mergeDensityControl(result,
                rovRecallRankNew, flowPoolRankNew, densityRules);

        return new RankResult(resultWithDensity);
    }

    public Map<String, Map<String, String>> getVideoRedisFeature(List<String> vids, String redisKeyPrefix) {
        List<String> keys = vids.stream().map(r -> redisKeyPrefix + r).collect(Collectors.toList());
        List<String> key2Values = this.redisTemplate.opsForValue().multiGet(keys);
        Map<String, Map<String, String>> result = new HashMap<>(vids.size());
        if (key2Values != null) {
            int j = 0;
            for (String vid : vids) {
                String vF = key2Values.get(j);
                ++j;
                if (vF == null) {
                    continue;
                }
                Map<String, String> vfMap = new HashMap<>();
                try {
                    vfMap = JSONUtils.fromJson(vF, new TypeToken<Map<String, String>>() {
                    }, vfMap);
                } catch (Exception e) {
                    log.error(String.format("parse video json is wrong on redisKeyPrefix in {} with {}", this.CLASS_NAME, redisKeyPrefix));
                }
                result.put(vid, vfMap);
            }
        }
        return result;
    }

    static class Tuple4 {
        public Map<String, String> first;
        public Map<String, String> second;
        public Map<String, String> third;

        public String name;

        public Tuple4(Map<String, String> first, Map<String, String> second, Map<String, String> third, String name) {
            this.first = first;
            this.second = second;
            this.third = third;
            this.name = name;
        }

    }

    static class Tuple3 {
        public String first;
        public List<String> second;
        public Map<String, String> third;

        public Tuple3(String first, List<String> second, Map<String, String> third) {
            this.first = first;
            this.second = second;
            this.third = third;
        }
    }

    static class Tuple2 {
        public Map<String, String> first;

        public String name;

        public Tuple2(Map<String, String> first, String name) {
            this.first = first;
            this.name = name;
        }

    }

    protected double restoreScore(double score) {
        return (0.1 * score) / (1 - 0.9 * score);
    }

    protected double restoreScore(double score, double w) {
        return (w * score) / (1 - (1 - w) * score);
    }

    Map<String, double[]> bucketsMap = new HashMap<>();
    Map<String, Double> bucketsLen = new HashMap<>();

    public void readBucketFile() {
        InputStream resourceStream = RankStrategy4RegionMergeModelBasic.class.getClassLoader().getResourceAsStream("20240609_bucket_274.txt");
        if (resourceStream != null) {
            try (BufferedReader reader = new BufferedReader(new InputStreamReader(resourceStream))) {
                Map<String, double[]> bucketsMap = new HashMap<>();
                Map<String, Double> bucketsLen = new HashMap<>();
                String line;
                while ((line = reader.readLine()) != null) {
                    // 替换空格和换行符，过滤空行
                    line = line.replace(" ", "").replaceAll("\n", "");
                    if (!line.isEmpty()) {
                        String[] rList = line.split("\t");
                        if (rList.length == 3) {
                            String key = rList[0];
                            double value1 = Double.parseDouble(rList[1]);
                            bucketsLen.put(key, value1);
                            double[] value2 = Arrays.stream(rList[2].split(","))
                                    .mapToDouble(Double::valueOf)
                                    .toArray();
                            bucketsMap.put(key, value2);
                        }
                    }
                }
                this.bucketsMap = bucketsMap;
                this.bucketsLen = bucketsLen;
            } catch (IOException e) {
                log.error("something is wrong in parse bucket file:" + e);
            }
        } else {
            log.error("no bucket file");
        }
    }

}
